import pandas as pd


class DataTranspose:

    @staticmethod
    def to_stack(df_data: pd.DataFrame, df_info: pd.DataFrame, dict_data_structure: dict) -> (pd.DataFrame, pd.DataFrame):

        """
        :param df_data: dataframe
        :param df_info: dataframe
        :param dict_data_structure: dict = {
            'id_col': str,
            'sp_col': str,
            'lst_scr': list,
            'dict_sp': dict = {
                1: {
                    'Ma_SP1': 'Ma_SP',
                    'Q1_SP1': 'Q1',
                    'Q2_SP1': 'Q2',
                    'Q3_SP1': 'Q3',
                },
                2: {
                    'Ma_SP2': 'Ma_SP',
                    'Q1_SP2': 'Q1',
                    'Q2_SP2': 'Q2',
                    'Q3_SP2': 'Q3',
                 },
            },
            'lst_fc': list
        }
        :return: df_data_stack, df_info_stack both as dataframe type
        """
        
        id_col = dict_data_structure['id_col']
        sp_col = dict_data_structure['sp_col']
        lst_scr = dict_data_structure['lst_scr']
        dict_sp = dict_data_structure['dict_sp']
        lst_fc = dict_data_structure['lst_fc']

        # df_data_stack generate
        df_data_scr = df_data.loc[:, [id_col] + lst_scr].copy()

        df_data_df = pd.DataFrame()
        if lst_fc:
            df_data_df = df_data.loc[:, [id_col] + lst_fc].copy()

        lst_df_data_sp = [df_data.loc[:, [id_col] + list(val.keys())].copy() for val in dict_sp.values()]

        for i, df in enumerate(lst_df_data_sp):
            df.rename(columns=dict_sp[i + 1], inplace=True)

        df_data_stack = pd.concat(lst_df_data_sp, axis=0, ignore_index=True)

        df_data_stack = df_data_scr.merge(df_data_stack, how='left', on=[id_col])

        if lst_fc:
            df_data_stack = df_data_stack.merge(df_data_df, how='left', on=[id_col])

        df_data_stack.reset_index(drop=True, inplace=True)

        df_data_stack.sort_values(by=[id_col, sp_col], inplace=True)
        df_data_stack.reset_index(drop=True, inplace=True)

        df_info_stack = df_info.copy()

        for key, val in dict_sp[1].items():
            df_info_stack.loc[df_info_stack['var_name'] == key, ['var_name']] = [val]

        df_info_stack.drop_duplicates(subset=['var_name'], keep='first', inplace=True)

        # Reset df_info_stack index
        df_info_stack['idx_var_name'] = df_info_stack['var_name']
        df_info_stack.set_index('idx_var_name', inplace=True)
        df_info_stack = df_info_stack.loc[list(df_data_stack.columns), :]
        df_info_stack.reindex(list(df_data_stack.columns))
        df_info_stack.reset_index(drop=True, inplace=True)

        return df_data_stack, df_info_stack


    @staticmethod
    def to_unstack(df_data_stack: pd.DataFrame, df_info_stack: pd.DataFrame, dict_unstack_structure: dict) -> (pd.DataFrame, pd.DataFrame):

        """
        :param df_data_stack: dataframe
        :param df_info_stack: dataframe
        :param dict_unstack_structure: dict = {
            'id_col': str,
            'sp_col': str,
            'lst_col_part_head': list,
            'lst_col_part_body': list,
            'lst_col_part_tail': list
        }
        :return: df_data_unstack, df_info_unstack both as dataframe type
        """

        id_col = dict_unstack_structure['id_col']
        sp_col = dict_unstack_structure['sp_col']

        dict_sp_val = df_info_stack.query(f"var_name == '{sp_col}'")['val_lbl'].values[0]

        df_part_head = df_data_stack.query(f"{sp_col} == {list(dict_sp_val.keys())[0]}")[[id_col] + dict_unstack_structure['lst_col_part_head']].copy()
        df_part_tail = df_data_stack.query(f"{sp_col} == {list(dict_sp_val.keys())[0]}")[[id_col] + dict_unstack_structure['lst_col_part_tail']].copy()

        df_part_body = pd.DataFrame()
        df_info_part_body = pd.DataFrame()

        dict_sort_col = dict()
        for k, v in dict_sp_val.items():

            sp_lbl = str(v).replace(' ', '_')

            df_info_by_k = df_info_stack.query(f"var_name.isin({dict_unstack_structure['lst_col_part_body']})").copy()
            df_data_by_k = df_data_stack.query(f"{sp_col} == {k}")[[id_col] + dict_unstack_structure['lst_col_part_body']].copy()

            dict_rename_col = dict()
            dict_rename_var_lbl = dict()
            for idx in df_info_by_k.index:

                var_name = df_info_by_k.at[idx, 'var_name']
                var_lbl = df_info_by_k.at[idx, 'var_lbl']
                var_type = df_info_by_k.at[idx, 'var_type']

                str_ma_name = str()

                if var_type in ['MA']:
                    str_ma_name, str_ma_cat = var_name.rsplit('_', 1)
                    str_var_name_new = f"{str_ma_name}_{sp_lbl}_{str_ma_cat}"
                else:
                    str_var_name_new = f"{var_name}_{sp_lbl}"

                dict_rename_col.update({var_name: str_var_name_new})
                dict_rename_var_lbl.update({var_lbl: f"{var_lbl}_{sp_lbl}"})

                if var_type in ['MA']:
                    if str_ma_name in dict_sort_col.keys():
                        dict_sort_col[str_ma_name].append(str_var_name_new)
                    else:
                        dict_sort_col.update({str_ma_name: [str_var_name_new]})
                else:
                    if var_name in dict_sort_col.keys():
                        dict_sort_col[var_name].append(str_var_name_new)
                    else:
                        dict_sort_col.update({var_name: [str_var_name_new]})

            df_data_by_k.rename(columns=dict_rename_col, inplace=True)
            df_info_by_k['var_name'].replace(dict_rename_col, inplace=True)
            df_info_by_k['var_lbl'].replace(dict_rename_var_lbl, inplace=True)

            if df_part_body.empty:
                df_part_body = df_data_by_k.copy()
                df_info_part_body = df_info_by_k.copy()
            else:
                df_part_body = df_part_body.merge(df_data_by_k, how='left', on=id_col)
                df_info_part_body = pd.concat([df_info_part_body, df_info_by_k], ignore_index=True)

        # Need to sort vars
        lst_sort_col = list()
        for v in dict_sort_col.values():
            lst_sort_col.extend(v)

        df_part_body = df_part_body.reindex(columns=[id_col] + lst_sort_col)

        df_info_part_body['idx_by_var_name'] = df_info_part_body['var_name']
        df_info_part_body.set_index('idx_by_var_name', inplace=True)
        df_info_part_body = df_info_part_body.reindex(lst_sort_col)
        df_info_part_body.reset_index(drop=True, inplace=True)
        # Need to sort vars

        df_data_unstack = df_part_head.copy()
        df_data_unstack = df_data_unstack.merge(df_part_body, how='left', on=id_col)
        df_data_unstack = df_data_unstack.merge(df_part_tail, how='left', on=id_col)

        df_info_unstack = df_info_stack.query(
            f"var_name.isin({[id_col] + dict_unstack_structure['lst_col_part_head']})").copy()
        df_info_unstack = pd.concat([df_info_unstack, df_info_part_body], ignore_index=True)
        df_info_unstack = pd.concat([df_info_unstack, df_info_stack.query(
            f"var_name.isin({dict_unstack_structure['lst_col_part_tail']})").copy()], ignore_index=True)

        df_data_unstack.reset_index(drop=True, inplace=True)
        df_info_unstack.reset_index(drop=True, inplace=True)

        return df_data_unstack, df_info_unstack


import numpy
import pandas as pd
import numpy as np


class CodeframeReader:

    def __init__(self, cf_file_name: str):
        self.cf_file_name = cf_file_name

        self.dict_add_new_qres_oe = dict()
        self.df_full_oe_coding = pd.DataFrame


    def to_dataframe_file(self):

        file_name = self.cf_file_name

        dict_df_ws = pd.read_excel(file_name, sheet_name=None, header=None)
        dict_df_coding = dict()
        dict_df_codelist = dict()

        df_rid = pd.DataFrame(data=[], columns=['RESPONDENTID'])
        for ws_name, df_ws in dict_df_ws.items():

            if ws_name in ['RAWDATA', 'DATABASE', 'VERBATIM']:
                continue

            if '-CODE' in ws_name:
                df_ws.columns = df_ws.iloc[1].tolist()
                df_ws = df_ws.query("index >= 2").copy()
                df_ws.reset_index(drop=True, inplace=True)
                df_rid = pd.concat([df_rid, df_ws.loc[:, ['RESPONDENTID']]], axis=0)
                dict_df_coding.update({ws_name: df_ws})
            else:
                qre_lbl = df_ws.at[0, 1]
                df_ws.columns = df_ws.iloc[2].tolist()
                df_ws = df_ws.query("index >= 3 & RECODE.isnull()").copy()
                df_ws.drop(columns=['RECODE', 'LABEL ENG'], inplace=True)
                df_ws.reset_index(drop=True, inplace=True)

                dict_df_codelist.update({
                    ws_name: {
                        'qre_lbl': qre_lbl,
                        'df_colname': pd.DataFrame(),
                        'df_codelist': df_ws
                    }
                })


        print('EXPORT CODING FILE')
        df_rid.drop_duplicates(subset=['RESPONDENTID'], inplace=True)
        df_rid.sort_values(by=['RESPONDENTID'], inplace=True)
        lst_rid = df_rid['RESPONDENTID'].values.tolist()

        lst_ws_col = ['Unnamed: 0', 'RESPONDENTID', 'COLUMN_NAME', 'VERBATIM', 'CODING', 'FW_CHECK']

        df_full_oe = pd.DataFrame()

        for ws_name, df_ws in dict_df_coding.items():

            print(ws_name)

            df_ws.rename(columns={'COLUMN NAME': 'COLUMN_NAME', 'FW CHECK': 'FW_CHECK'}, inplace=True)

            lst_qre_comb = list(dict.fromkeys([a.replace('Y1_', 'Y1|').replace('Y2_', 'Y2|').rsplit('|', 1)[0] for a in df_ws['COLUMN_NAME'].values.tolist()]))

            df_ws['VERBATIM'] = df_ws['VERBATIM'].astype(str)
            df_ws['CODING'] = df_ws['CODING'].astype(str)

            df_ws_new = pd.DataFrame(columns=lst_ws_col, data=[])
            for rid in lst_rid:
                for qre in lst_qre_comb:

                    df_fil = df_ws.query(f"RESPONDENTID == '{rid}' & COLUMN_NAME.str.contains('{qre}')")

                    verbatim = '|'.join(df_fil['VERBATIM'].values.tolist())
                    coding = '\\'.join(df_fil['CODING'].values.tolist())
                    lst_coding = list(dict.fromkeys(coding.split('\\')))

                    if lst_coding[0] == '' and 'THICHHON' not in str(ws_name).upper() and 'THICH_HON' not in str(ws_name).upper():
                        lst_coding[0] = '99999'

                    if len(lst_coding) > 1 and '99999' in lst_coding:
                        lst_coding.remove('99999')

                    df_check = df_ws_new.query(f"RESPONDENTID == '{rid}' & COLUMN_NAME == '{qre}'")

                    if not df_check.empty:
                        df_ws_new.loc[((df_ws_new['RESPONDENTID'] == rid) & (df_ws_new['COLUMN_NAME'] == qre)), :] = [f"{rid}@_@{qre}", rid, qre, verbatim, lst_coding, np.nan]
                    else:
                        df_ws_new = pd.concat([df_ws_new, pd.DataFrame(
                            columns=lst_ws_col,
                            data=[[f"{rid}@_@{qre}", rid, qre, verbatim, lst_coding, np.nan]])], axis=0)


            df_ws_new['VERBATIM'].replace({'': 'NONE'}, inplace=True)

            df_ws_new.reset_index(drop=True, inplace=True)

            for rid in lst_rid:
                df_fil = df_ws_new.query(f"RESPONDENTID == '{rid}' & COLUMN_NAME.isin({lst_qre_comb})")
                qre_total_name = lst_qre_comb[0].replace('_Y1', '_Total')

                verbatim = '|'.join(df_fil['VERBATIM'].values.tolist())

                lst_coding = list()
                for item in df_fil['CODING'].values.tolist():
                    lst_coding.extend(item)

                lst_coding = list(dict.fromkeys(lst_coding))

                if len(lst_coding) > 1 and '99999' in lst_coding:
                    lst_coding.remove('99999')

                lst_total_row = [f"{rid}@_@{qre_total_name}", rid, qre_total_name, verbatim, lst_coding, np.nan]

                df_ws_new = pd.concat([df_ws_new, pd.DataFrame(columns=lst_ws_col, data=[lst_total_row])], axis=0)

            df_ws_new.sort_values(by=['RESPONDENTID', 'COLUMN_NAME'], inplace=True)
            df_ws_new.reset_index(drop=True, inplace=True)

            df_ws_new['CODING_LEN'] = [len(a) for a in df_ws_new['CODING']]
            df_ws_new['CODING'] = ['\\'.join(a) for a in df_ws_new['CODING']]
            max_len = df_ws_new['CODING_LEN'].max()
            arr_max_len = [a for a in range(1, max_len + 1)]

            df_ws_new[arr_max_len] = df_ws_new['CODING'].str.split('\\', expand=True)

            df_ws_new = pd.melt(df_ws_new, id_vars=['RESPONDENTID', 'COLUMN_NAME'], value_vars=arr_max_len)
            df_ws_new.sort_values(by=['RESPONDENTID', 'COLUMN_NAME'], inplace=True)
            df_ws_new.reset_index(drop=True, inplace=True)

            # df_ws_new['value'].replace({None: np.nan}, inplace=True)
            df_ws_new['value'] = df_ws_new['value'].astype(float)

            df_ws_new['COLUMN_NAME'] = [f"{a1}_OE_{a2}" for a1, a2 in zip(df_ws_new['COLUMN_NAME'], df_ws_new['variable'])]

            lst_qre_name = df_ws_new['COLUMN_NAME'].copy().drop_duplicates(keep='first').values.tolist()

            df_ws_new.drop_duplicates(subset=['RESPONDENTID', 'COLUMN_NAME', 'variable'], inplace=True)

            df_ws_new = df_ws_new.set_index(['RESPONDENTID', 'COLUMN_NAME'])['value'].unstack().reset_index()

            df_ws_new = df_ws_new.reindex(columns=['RESPONDENTID'] + lst_qre_name)

            print(f'ADD columns to dict_df_codelist[{ws_name.replace("-CODE", "")}]')
            lst_qre = list(df_ws_new.columns)
            lst_qre.remove('RESPONDENTID')
            df_colname = pd.DataFrame(columns=['COL_NAME'], data=lst_qre)
            df_colname[['COL_NAME', 'stt']] = df_colname['COL_NAME'].str.rsplit('_', n=1, expand=True)
            df_colname.drop_duplicates(subset=['COL_NAME'], keep='last', inplace=True)
            df_colname['COL_NAME'] = df_colname['COL_NAME'] + '|' + df_colname['stt'].astype(str)
            df_colname.drop(columns='stt', inplace=True)
            df_colname.reset_index(drop=True, inplace=True)
            df_colname = pd.concat([df_colname, pd.DataFrame(columns=['SEC', 'LABEL', 'TYPE', 'CODELIST'], data=[['PRODUCT|FORCE_CHOICE|NORMAL_OE', '', 'MA', {}]] * df_colname.shape[0])], axis=1)
            df_colname['LABEL'] = dict_df_codelist[ws_name.replace('-CODE', '')]['qre_lbl']
            dict_df_codelist[ws_name.replace('-CODE', '')].update({'df_colname': df_colname})

            if df_full_oe.empty:
                df_full_oe = df_ws_new.copy()
            else:
                df_full_oe = df_full_oe.merge(df_ws_new, how='left', on='RESPONDENTID')


        # df_full_oe.to_csv(f"{file_name}_CODING.csv", index=False)

        print('EXPORT CODELIST FILE')
        df_full_codelist = pd.DataFrame()
        for ws_name, val in dict_df_codelist.items():
            print(ws_name)
            dict_codelist = dict()
            net_count = 900001
            df_codelist = val['df_codelist']
            cur_net_key = str()
            for idx in df_codelist.index:
                code = df_codelist.at[idx, 'CODE']
                label = df_codelist.at[idx, 'LABEL VNI']

                if pd.isnull(df_codelist.at[idx, 'CODE']):
                    notes = df_codelist.at[idx, 'NOTES']
                    cur_net_key = f"{net_count}|{notes}|{label}"
                    dict_codelist.update({cur_net_key: {}})
                    net_count += 1
                else:
                    if cur_net_key:
                        dict_codelist[cur_net_key].update({str(int(code)): label})
                    else:
                        dict_codelist.update({str(int(code)): label})

            df_colname = val['df_colname']
            df_colname['CODELIST'] = [{'net_code': dict_codelist}] * df_colname.shape[0]

            if df_full_codelist.empty:
                df_full_codelist = df_colname.copy()
            else:
                df_full_codelist = pd.concat([df_full_codelist, df_colname], axis=0)

        df_full_codelist.reset_index(drop=True, inplace=True)
        # df_full_codelist.to_csv(f"{file_name}_CODELIST.csv", index=False, encoding='utf-8-sig')

        df_full_codelist.drop(columns=['SEC'], inplace=True)

        with pd.ExcelWriter(f'{file_name}_output.xlsx', engine="openpyxl") as writer:
            df_full_codelist.to_excel(writer, sheet_name='codelist')
            df_full_oe.to_excel(writer, sheet_name='coding')

        df_full_codelist.set_index('COL_NAME', drop=True, inplace=True)
        df_full_codelist.loc[:, 'VALUES'] = np.nan

        dict_add_new_qres_oe = df_full_codelist.to_dict('index')

        for k, v in dict_add_new_qres_oe.items():
            self.dict_add_new_qres_oe.update({k: list(v.values())})

        self.df_full_oe_coding = df_full_oe


import pandas as pd
import numpy as np

# from dpkits.ap_data_converter import APDataConverter
# from dpkits.data_processing import DataProcessing
# from dpkits.data_transpose import DataTranspose
# from dpkits.table_generator import DataTableGenerator
# from dpkits.table_formater import TableFormatter
# from dpkits.calculate_lsm import LSMCalculation


# IGNORE THIS-----------------------------------------------------------------------------------------------------------
import sys
sys.path.insert(0, 'C:/Users/PC/OneDrive/Dev Area/PyPackages/packaging_dpkits/src/dpkits')

from ap_data_converter import APDataConverter
from data_processing import DataProcessing
from data_transpose import DataTranspose
from table_generator import DataTableGenerator
from table_formater import TableFormatter
from codeframe_reader import CodeframeReader
from calculate_lsm import LSMCalculation
# IGNORE THIS-----------------------------------------------------------------------------------------------------------



# Define input/output files name
str_file_name = 'VN8413_Data'
str_tbl_file_name = f'{str_file_name}_Topline.xlsx'


# Call Class APDataConverter with file_name
converter = APDataConverter(file_name=f'{str_file_name}.xlsx')

converter.lstDrop.extend(['DV'])

"""
README: Convert input file to dataframe

- df_data: contains data as pandas dataframe
- df_info: contains data info as pandas dataframe (ex: var_name, var_lbl, var_type, val_lbl)
    - var_name = data column name (variable)
    - var_lbl = variable label
    - var_type = variable type
    - val_lbl = value label
"""


df_data, df_info = converter.convert_df_mc()  # Use 'converter.convert_df_md()' if you need md data

# LSM 6 CALCULATION - Only use for Unilever projects which have LSM questions
# df_data, df_info = LSMCalculation.cal_lsm_6(df_data, df_info)

df_data = pd.DataFrame(df_data)
df_info = pd.DataFrame(df_info)

# AFTER CONVERTING YOU CAN DO ANYTHING WITH DATAFRAME-------------------------------------------------------------------
# df_info columns must be ['var_name', 'var_lbl', 'var_type', 'val_lbl']

dict_add_new_qres = {
    'Ma_SP_1': ['Mã SP', 'SA', {'1': 'Concept 1', '2': 'Concept 2', '3': 'Concept 3'}, 1],
    'Ma_SP_2': ['Mã SP', 'SA', {'1': 'Concept 1', '2': 'Concept 2', '3': 'Concept 3'}, 2],
    'Ma_SP_3': ['Mã SP', 'SA', {'1': 'Concept 1', '2': 'Concept 2', '3': 'Concept 3'}, 3],

    'New_FT': ['New FT', 'FT', {}, np.nan],
    'New_Num': ['New Num', 'NUM', {}, np.nan],
    'New_MA|6': ['New MA', 'MA', {'1': '1', '2': '2', '3': '3', '4': '4', '5': '5', '6': '6', '7': '7'}, np.nan],
}

df_data, df_info = DataProcessing.add_qres(df_data, df_info, dict_add_new_qres)
df_data = pd.DataFrame(df_data)
df_info = pd.DataFrame(df_info)







# Just for checking
with pd.ExcelWriter(f'{str_file_name}_preview.xlsx', engine="openpyxl") as writer:
    df_data.to_excel(writer, sheet_name='df_data')
    df_info.to_excel(writer, sheet_name='df_info')


# TRANSPOSE TO STACK----------------------------------------------------------------------------------------------------
lst_scr = [
    'S1',
    'S2',
    'S3_a',
    'S3_b',
    'S4',
    'S5',
    'S6_1',
    'S6_2',
    'S6_3',
    'S6_4',
    'S6_5',
    'S6_6',
    'S7',
    'S8',
    'S10',
]

lst_fc = ['F1_ByProdCode']

dict_stack_structure = {
    'id_col': 'ID',
    'sp_col': 'Ma_SP',
    'lst_scr': lst_scr,
    'dict_sp': {
        1: {
            'Ma_SP_1': 'Ma_SP',
            'Q1_1': 'Q1',
            'Q2_1': 'Q2',
            'Q3_1': 'Q3',
            'Q4_1': 'Q4',
            'Q5_1': 'Q5',
            'Q9_1': 'Q9',
            'Q6_1': 'Q6',
            'Q7_1': 'Q7',
            'Q8_1': 'Q8',
            'Q10_1': 'Q10',
            'F1_YN_1': 'F1_YN_New',
            'F2_OE_1': 'F2_OE_New',
        },
        2: {
            'Ma_SP_2': 'Ma_SP',
            'Q1_2': 'Q1',
            'Q2_2': 'Q2',
            'Q3_2': 'Q3',
            'Q4_2': 'Q4',
            'Q5_2': 'Q5',
            'Q9_2': 'Q9',
            'Q6_2': 'Q6',
            'Q7_2': 'Q7',
            'Q8_2': 'Q8',
            'Q10_2': 'Q10',
            'F1_YN_2': 'F1_YN_New',
            'F2_OE_2': 'F2_OE_New',
        },
        3: {
            'Ma_SP_3': 'Ma_SP',
            'Q1_3': 'Q1',
            'Q2_3': 'Q2',
            'Q3_3': 'Q3',
            'Q4_3': 'Q4',
            'Q5_3': 'Q5',
            'Q9_3': 'Q9',
            'Q6_3': 'Q6',
            'Q7_3': 'Q7',
            'Q8_3': 'Q8',
            'Q10_3': 'Q10',
            'F1_YN_3': 'F1_YN_New',
            'F2_OE_3': 'F2_OE_New',
        },
    },
    'lst_fc': lst_fc
}

df_data_stack, df_info_stack = DataTranspose.to_stack(df_data, df_info, dict_stack_structure)




# TRANSPOSE TO UNSTACK--------------------------------------------------------------------------------------------------
dict_unstack_structure = {
    'id_col': 'ID',
    'sp_col': 'Ma_SP',
    'lst_col_part_head': lst_fc,
    'lst_col_part_body': [
        'Ma_SP',
        'Q1',
        'Q2',
        'Q3',
        'Q4',
        'Q5',
        'Q9',
        'Q6',
        'Q7',
        'Q8',
        'Q10',
        'F1_YN_New',
        'F2_OE_New',
    ],
    'lst_col_part_tail': lst_fc
}

df_data_unstack, df_info_unstack = DataTranspose.to_unstack(df_data_stack, df_info_stack, dict_unstack_structure)


# # ----------------------------------------------------------------------------------------------------------------------
# # OE RUNNING------------------------------------------------------------------------------------------------------------
# # ----------------------------------------------------------------------------------------------------------------------
# # NOT YET START
#
# cfr = CodeframeReader(cf_file_name='VN8413_Codeframe.xlsm')
#
# cfr.to_dataframe_file()
#
# df_data_stack, df_info_stack = DataProcessing.add_qres(df_data_stack, df_info_stack, cfr.dict_add_new_qres_oe)
# df_data_stack, df_info_stack = pd.DataFrame(df_data_stack), pd.DataFrame(df_info_stack)
#
# df_coding = pd.DataFrame(cfr.df_full_oe_coding)
#
# df_coding[['ID', 'Ma_SP']] = df_coding['RESPONDENTID'].str.rsplit('_', n=1, expand=True)
# df_coding.drop(columns=['RESPONDENTID'], inplace=True)
#
# df_data_stack['Ma_SP'] = df_data_stack['Ma_SP'].astype(int)
# df_coding['Ma_SP'] = df_coding['Ma_SP'].astype(int)
#
# lst_drop = df_coding.columns.tolist()
# lst_drop.remove('ID')
# lst_drop.remove('Ma_SP')
#
# df_data_stack.drop(columns=lst_drop, inplace=True)
#
# df_data_stack = df_data_stack.merge(df_coding, how='left', on=['ID', 'Ma_SP'])
#
# for i in lst_drop:
#     df_data_stack[i].replace({99999: np.nan}, inplace=True)



# ----------------------------------------------------------------------------------------------------------------------
# EXPORT SAV DATA FILES-------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------
dict_dfs = {
    1: {
        'data': df_data,
        'info': df_info,
        'tail_name': 'ByCode',
        'sheet_name': 'ByCode',
        'is_recode_to_lbl': False,
    },
    2: {
        'data': df_data_stack,
        'info': df_info_stack,
        'tail_name': 'Stack',
        'sheet_name': 'Stack',
        'is_recode_to_lbl': False,
    },
    3: {
        'data': df_data_unstack,
        'info': df_info_unstack,
        'tail_name': 'Unstack',
        'sheet_name': 'Unstack',
        'is_recode_to_lbl': False,
    },
}

converter.generate_multiple_data_files(dict_dfs=dict_dfs, is_zip=False, is_export_sav=False)


# ----------------------------------------------------------------------------------------------------------------------
# EXPORT DATA TABLES----------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------
"""
README:
    - Side question properties:
    {
        "qre_name": "$Q15",  # column name, must set '$' if it is MA question
        "qre_lbl": "{lbl}: new label",  # default df_info label, input {lbl} top keep original label and addin new label
        "qre_filter": "Age.isin([2, 3])",  # use for filter question
        "sort": "des", # sort options: acs / des
        "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}  # calculate mean base on dict: key == code in data, value = weighted values
        "cats": {  # use for define net/combine code with specify format
            'net_code': {
                '900001|combine|Group 1 + 2': {
                    '1': 'Yellow/dull teeth',
                    '2': 'Sensitive teeth',
                    '3': 'Dental plaque',
                    '4': 'Caries',
                },
                '900002|net|Group 1': {
                    '1': 'Yellow/dull teeth',
                    '2': 'Sensitive teeth',
                },
                '900003|net|Group 2': {
                    '3': 'Dental plaque',
                    '4': 'Caries',
                },
            },
            '8': 'Other (specify)',
            '9': 'No problem',
        }
    },

    - Header question properties:
    {
        "qre_name": "S1",  # define 'S1' if SA, '$S1' if MA, '@S1_xxx' if create header base on specify condition
        "qre_lbl": "City",  # typing every label is fine
        "cats":
            # SA/MA: define category list base on df_info, use 'TOTAL' if need to display total column
            {
                "TOTAL": "TOTAL",
                '3': 'Hồ Chí Minh',
                '4': 'Cần Thơ'

            }
            # @: define header base on specify condition
            {
                "S3_b.isin([2])": "<=30 (22-30 tuổi)",
                "S3_b.isin([3, 4])": ">30 (31-39 tuổi)",
            }
    },

    - Table properties:
        + key: table key name
        + value: table specify properties

        "Main": {
            "tbl_name": "Main",  # display on excel sheet name
            "tbl_filter": "Ma_SP > 0",  # filter of this table
            "is_count": 0,  # 1 for count, 0 for percentage
            "is_pct_sign": 1,  # 1 for display '%' else 0
            "is_hide_oe_zero_cats": 1,  # 1 for hide answers which percentage = 0% at all header columns
            "is_hide_zero_cols": 1,  # 1 for hide header columns which percentage = 0% at all row
            "sig_test_info":  # define significant test
            {
                "sig_type": "rel",  # 'rel' for dependent sig test, 'ind' for independent sig test
                "sig_cols": [],  # define columns to sig, leave it blank if need to sig all columns
                "lst_sig_lvl": [90, 95]  # sig level: maximum 2 levels
            },
            "lst_side_qres": lst_side_main,  # list of side question
            "lst_header_qres": lst_header  # list of header defines
        },

"""


lst_header = [
    # header lvl 1
    [
        {
            "qre_name": "S1",
            "qre_lbl": "City",
            "cats": {
                "TOTAL": "TOTAL",
                '3': 'Hồ Chí Minh',
                '4': 'Cần Thơ'

            }
        },
    ],

    # header lvl 2
    [
        {
            "qre_name": "@S3_b_Group",
            "qre_lbl": "Age",
            "cats": {
                "S3_b > 0": "TOTAL",
                "S3_b.isin([2])": "<=30 (22-30 tuổi)",
                "S3_b.isin([3, 4])": ">30 (31-39 tuổi)",
            }
        },
        {
            "qre_name": "@S4_Class",
            "qre_lbl": "Class",
            "cats": {
                "S4.isin([1, 2])": "A&B (Từ 13,500,000 đến 22,499,000 VND & Trên 22,500,000)",
                "S4.isin([3])": "C (Từ 7,500,000 đến 13,499,000 VND)",
            }
        },
        {
            "qre_name": "@S8_BUMO",
            "qre_lbl": "BUMO",
            "cats": {
                "S8.isin([2])": "Tiger nâu",
                "S8.isin([6, 7, 8])": "Sài Gòn",
                "S8.isin([12, 13, 14])": "Larue",
            }
        },
    ],

    # header lvl 3
    [
        {
            "qre_name": "Ma_SP",
            "qre_lbl": "Mã Concept",
            "cats": {'1': 'Concept 1', '2': 'Concept 2', '3': 'Concept 3'}
        },
    ],
]

lst_side_scr_tagon = [
    {"qre_name": "S1"},
    {"qre_name": "S2"},
    {"qre_name": "S3_b"},
    {"qre_name": "S4"},
    {"qre_name": "S5"},
    {"qre_name": "$S6"},
    {"qre_name": "S7"},
    {"qre_name": "S8"},
    {"qre_name": "S10"},

    {"qre_name": "Dealer_HCM_01_Rank1"},
    {"qre_name": "$Dealer_HCM_02_Rank"},


]



lst_side_main = [
    {"qre_name": "Q1", 'cats': {
        '1': '1 - Hoàn toàn không thích', '2': '2 - Không thích', '3': '3 - Không thích cũng không ghét', '4': '4 - Thích', '5': '5 - Rất thích',
        'net_code': {
            '900001|combine|T2B': {'4': '4', '5': '5'},
            '900002|combine|Medium': {'3': '3'},
            '900003|combine|B2B': {'1': '1', '2': '2'},
        }
    }, "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}},

    {"qre_name": "Q4", 'cats': {
        '1': 'Hoàn toàn không phù hợp', '2': 'Không phù hợp', '3': 'Hơi không phù hợp', '4': 'Phù hợp', '5': 'Rất Phù hợp',
        'net_code': {
            '900001|combine|T2B': {'4': '4', '5': '5'},
            '900002|combine|Medium': {'3': '3'},
            '900003|combine|B2B': {'1': '1', '2': '2'},
        }
    }, "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}},

    {"qre_name": "Q5", 'cats': {
        '1': 'Hoàn toàn không mới lạ và khác biệt', '2': 'Không mới lạ và khác biệt', '3': 'Hơi không mới lạ và khác biệt', '4': 'Mới lạ và khác biệt', '5': 'Rất mới lạ và khác biệt',
        'net_code': {
            '900001|combine|T2B': {'4': '4', '5': '5'},
            '900002|combine|Medium': {'3': '3'},
            '900003|combine|B2B': {'1': '1', '2': '2'},
        }
    }, "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}},

    {"qre_name": "Q9", 'cats': {
        '1': 'Hoàn toàn không cao cấp', '2': 'Không cao cấp', '3': 'Hơi không cao cấp', '4': 'Cao cấp', '5': 'Rất cao cấp',
        'net_code': {
            '900001|combine|T2B': {'4': '4', '5': '5'},
            '900002|combine|Medium': {'3': '3'},
            '900003|combine|B2B': {'1': '1', '2': '2'},
        }
    }, "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}},

    {"qre_name": "Q6", 'cats': {
        '1': 'Chắc chắn sẽ không mua', '2': 'Không mua', '3': 'Có thể sẽ mua hoặc không', '4': 'Sẽ mua', '5': 'Chắc chắn sẽ mua',
        'net_code': {
            '900001|combine|T2B': {'4': '4', '5': '5'},
            '900002|combine|Medium': {'3': '3'},
            '900003|combine|B2B': {'1': '1', '2': '2'},
        }
    }, "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}},

    {"qre_name": "Q10", 'cats': {
        '1': 'Chắc chắn sẽ không mua', '2': 'Không mua', '3': 'Có thể sẽ mua hoặc không', '4': 'Sẽ mua', '5': 'Chắc chắn sẽ mua',
        'net_code': {
            '900001|combine|T2B': {'4': '4', '5': '5'},
            '900002|combine|Medium': {'3': '3'},
            '900003|combine|B2B': {'1': '1', '2': '2'},
        },
    }, "mean": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}},

    {"qre_name": "F1_YN_New"},

]

lst_side_oe = [
    {"qre_name": "$Q2_OE"},
    {"qre_name": "$Q3_OE"},
    {"qre_name": "$Q7_OE"},
    {"qre_name": "$Q8_OE"},
    {"qre_name": "$F2_OE_OE"},
]


lst_func_to_run = [
    # SCREENER
    {
        'func_name': 'run_standard_table_sig',
        'tables_to_run': [
            'Scr_Tagon',
        ],
        'tables_format': {
            "Scr_Tagon": {
                "tbl_name": "Screener",
                "tbl_filter": "S1 > 0",
                "is_count": 1,
                "is_pct_sign": 0,
                "is_hide_oe_zero_cats": 1,
                "is_hide_zero_cols": 1,
                "sig_test_info": {
                    "sig_type": "",
                    "sig_cols": [],
                    "lst_sig_lvl": []
                },
                "lst_side_qres": lst_side_scr_tagon,
                "lst_header_qres": lst_header[:-1]
            },
        },
    },

    # MAIN
    {
        'func_name': 'run_standard_table_sig',
        'tables_to_run': [
            'Main',
            'Main_oe',
        ],
        'tables_format': {

            "Main": {
                "tbl_name": "Main",
                "tbl_filter": "Ma_SP > 0",
                "is_count": 0,
                "is_pct_sign": 1,
                "is_hide_oe_zero_cats": 1,
                "is_hide_zero_cols": 1,
                "sig_test_info": {
                    "sig_type": "rel",
                    "sig_cols": [],
                    "lst_sig_lvl": [90, 95]
                },
                "lst_side_qres": lst_side_main,
                "lst_header_qres": lst_header
            },

            "Main_oe": {
                "tbl_name": "Main_oe",
                "tbl_filter": "Ma_SP > 0",
                "is_count": 0,
                "is_pct_sign": 1,
                "is_hide_oe_zero_cats": 1,
                "is_hide_zero_cols": 1,
                "sig_test_info": {
                    "sig_type": "",
                    "sig_cols": [],
                    "lst_sig_lvl": []
                },
                "lst_side_qres": lst_side_oe,
                "lst_header_qres": lst_header
            },
        },

    },
]

# RUN TABLE FOR SCREENER
dtg = DataTableGenerator(df_data=df_data, df_info=df_info, xlsx_name=str_tbl_file_name)
dtg.run_tables_by_js_files(lst_func_to_run[:1])

# # RUN TABLE FOR MAIN
# dtg = DataTableGenerator(df_data=df_data_stack, df_info=df_info_stack, xlsx_name=str_tbl_file_name)
# dtg.run_tables_by_js_files(lst_func_to_run[1:], is_append=True)




# FORMAT TABLES---------------------------------------------------------------------------------------------------------
dtf = TableFormatter(xlsx_name=str_tbl_file_name)
dtf.format_sig_table()



print('\n==>TESTING PROCESS DONE')




import pandas as pd
import numpy as np



class DataProcessing:

    @staticmethod
    def add_qres(df_data: pd.DataFrame, df_info: pd.DataFrame, dict_add_new_qres: dict) -> (pd.DataFrame, pd.DataFrame):
        info_col_name = ['var_name', 'var_lbl', 'var_type', 'val_lbl']

        for key, val in dict_add_new_qres.items():

            if val[1] in ['MA']:
                qre_ma_name, max_col = str(key).rsplit('|', 1)

                for i in range(1, int(max_col) + 1):
                    df_info = pd.concat([df_info, pd.DataFrame(columns=info_col_name, data=[[f'{qre_ma_name}_{i}', val[0], val[1], val[2]]])], axis=0, ignore_index=True)
                    df_data = pd.concat([df_data, pd.DataFrame(columns=[f'{qre_ma_name}_{i}'], data=[val[-1]] * df_data.shape[0])], axis=1)

            else:
                df_info = pd.concat([df_info, pd.DataFrame(columns=info_col_name, data=[[key, val[0], val[1], val[2]]])], axis=0, ignore_index=True)
                df_data = pd.concat([df_data, pd.DataFrame(columns=[key], data=[val[-1]] * df_data.shape[0])], axis=1)

        df_data.reset_index(drop=True, inplace=True)
        df_info.reset_index(drop=True, inplace=True)

        return df_data, df_info




import openpyxl
from openpyxl.styles.borders import Border, Side
from openpyxl.styles import Alignment, Font
from openpyxl.styles.fills import PatternFill
from openpyxl.utils import get_column_letter



class TableFormatter:

    def __init__(self, xlsx_name):
        self.xlsx_name = xlsx_name

        self.thin_border = Border(left=Side(style='thin'), right=Side(style='thin'), top=Side(style='thin'), bottom=Side(style='thin'))
        self.dot_border = Border(left=Side(style='dotted'), right=Side(style='dotted'), top=Side(style='dotted'), bottom=Side(style='dotted'))

        self.dot_thin_right_border = Border(left=Side(style='dotted'), right=Side(style='thin'), top=Side(style='dotted'), bottom=Side(style='dotted'))
        self.dot_thin_bot_border = Border(left=Side(style='dotted'), right=Side(style='dotted'), top=Side(style='dotted'), bottom=Side(style='thin'))
        self.dot_right_thin_right_border = Border(left=Side(style='dotted'), right=Side(style='thin'), top=Side(style='dotted'), bottom=Side(style='thin'))

        self.medium_left_1_border = Border(left=Side(style='medium'), right=Side(style='dotted'), top=Side(style='dotted'), bottom=Side(style='dotted'))
        self.medium_left_2_border = Border(left=Side(style='medium'), right=Side(style='dotted'), top=Side(style='dotted'), bottom=Side(style='thin'))
        self.medium_left_3_border = Border(left=Side(style='medium'), right=Side(style='dotted'), top=Side(style='thin'), bottom=Side(style='dotted'))
        self.medium_left_4_border = Border(left=Side(style='medium'), right=Side(style='thin'), top=Side(style='thin'), bottom=Side(style='thin'))


    # def format_table(self):
    #
    #     wb = openpyxl.load_workbook(self.xlsx_name, data_only=True)
    #
    #     ws_content = wb['Content']
    #     ws_content.column_dimensions['B'].width = 60
    #
    #     for ws in wb.worksheets:
    #
    #         lst_sub_header_col = list()
    #
    #         if ws.title in ['Content']:
    #             continue
    #
    #         self.logger.info('Formatting %s' % ws.title)
    #
    #         if ws.title in ['Contribution']:
    #             self.format_contribution_sheet(ws, ws_content)
    #             continue
    #
    #         if ws.title in ['BRAND MASTER', 'TV-TOTAL', 'TV-OWNER', 'TV-INTENDER']:
    #             self.format_brandmaster_sheet(ws, ws_content)
    #             continue
    #
    #         self.format_content_sheet(ws_content, ws, str_cell_content_addr='E2')
    #
    #         is_matrix_table = False
    #         step = 0
    #         if ws.cell(2, 1).value == 'matrix_table':
    #             is_matrix_table = True
    #             step = int(ws.cell(2, 2).value)
    #
    #         last_header_row = 0
    #         for i in range(1, 6):
    #             if ws.cell(i + 1, 5).value == 'Base':
    #                 last_header_row = i
    #                 break
    #
    #         ws.freeze_panes = f'F{last_header_row + 1}'
    #
    #         num_format = ws.cell(3, 1).value
    #
    #         ws.column_dimensions['A'].width = 20
    #         ws.column_dimensions['B'].width = 35
    #         ws.column_dimensions['E'].width = 40
    #
    #         ws.column_dimensions['C'].hidden = True
    #         ws.column_dimensions['D'].hidden = True
    #
    #         ws.row_dimensions[1].hidden = True
    #
    #         if not is_matrix_table:
    #             for i in range(2, last_header_row + 1):
    #                 ws.row_dimensions[i].height = 30
    #
    #         if is_matrix_table:
    #             ws.row_dimensions[4].hidden = True
    #
    #         self.logger.info('Formatting %s - Header' % ws.title)
    #
    #         ws_max_col = ws.max_column
    #         dict_header_color = {
    #             0: {'fontColor': 'FFFFFF', 'fgColor': '538DD5'},
    #             1: {'fontColor': 'FFFFFF', 'fgColor': '8DB4E2'},
    #             2: {'fontColor': '1F497D', 'fgColor': 'C5D9F1'},
    #             3: {'fontColor': '1F497D', 'fgColor': 'DCE6F1'},
    #             4: {'fontColor': '1F497D', 'fgColor': 'DCE6F1'},
    #         }
    #
    #         for irow in range(2, last_header_row + 1):
    #             start_column = 6
    #
    #             for icol in range(6, ws_max_col + 1):
    #
    #                 cur_cell = ws.cell(irow, icol)
    #
    #                 cur_cell.border = self.thin_border
    #                 cur_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
    #                 cur_cell.font = Font(bold=True, size=12, color=dict_header_color[irow - 2]['fontColor'])
    #                 cur_cell.fill = PatternFill(patternType='solid', fgColor=dict_header_color[irow - 2]['fgColor'])
    #
    #                 if cur_cell.value != ws.cell(irow, icol + 1).value:
    #                     end_column = icol
    #
    #                     if start_column != end_column and cur_cell.value is not None:
    #
    #                         ws.merge_cells(start_row=irow, start_column=start_column, end_row=irow, end_column=end_column)
    #                         lst_sub_header_col.append(icol)
    #
    #                     start_column = icol + 1
    #
    #
    #                 if ws.cell(irow + 1, icol).value is None:
    #                     iirow = irow + 1
    #                     while not ws.cell(iirow, icol).value:
    #                         iirow += 1
    #
    #                     ws.merge_cells(start_row=irow, start_column=icol, end_row=iirow - 1, end_column=icol)
    #
    #
    #         # start_column = 6
    #         # for icol in range(6, ws.max_column + 1):
    #         #     cur_cell = ws.cell(2, icol)
    #         #     cur_cell.border = self.thin_border
    #         #     cur_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
    #         #     cur_cell.font = Font(bold=True, size=12, color='FFFFFF')
    #         #     cur_cell.fill = PatternFill(patternType='solid', fgColor='538DD5')
    #         #
    #         #     if cur_cell.value != ws.cell(2, icol + 1).value:
    #         #         end_column = icol
    #         #         ws.merge_cells(start_row=2, start_column=start_column, end_row=2, end_column=end_column)
    #         #         start_column = icol + 1
    #         #         lst_sub_header_col.append(icol)
    #
    #
    #         # start_column = 6
    #         # for icol in range(6, ws.max_column):
    #         #     cur_cell = ws.cell(3, icol)
    #         #     cur_cell.border = self.thin_border
    #         #     cur_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
    #         #     cur_cell.font = Font(bold=True, size=12)
    #         #     cur_cell.fill = PatternFill(patternType='solid', fgColor='D9E1F2')
    #         #
    #         #     if is_matrix_table and cur_cell.value != ws.cell(3, icol + 1).value:
    #         #         end_column = icol
    #         #         ws.merge_cells(start_row=3, start_column=start_column, end_row=3, end_column=end_column)
    #         #         start_column = icol + 1
    #         #         lst_sub_header_col.append(icol)
    #
    #
    #         self.logger.info('Formatting %s - Side axis' % ws.title)
    #         start_side_row = last_header_row + 1
    #         self.format_side_axis(ws, is_matrix_table, start_side_row, num_format, step, lst_sub_header_col)
    #
    #         if num_format in ['pct', 'pct_sign']:
    #             ws.cell(3, 1).value = 'percentage'
    #
    #     # output_name = self.xlsx_name.replace('.xlsx', '_output.xlsx')
    #     output_name = self.xlsx_name
    #
    #     self.logger.info('save wb as %s' % output_name)
    #     wb.save(output_name)
    #     wb.close()


    @staticmethod
    def format_content_sheet(ws_content, ws_dtbl, str_cell_content_addr):

        idx = ws_content.max_row + 1

        ws_dtbl[str_cell_content_addr].value = 'Content'
        ws_dtbl[str_cell_content_addr].hyperlink = f"#'Content'!B{idx}"
        ws_dtbl[str_cell_content_addr].font = Font(color='0000FF')

        ws_content[f'A{idx}'].value = ws_content.max_row
        ws_content[f'B{idx}'].value = ws_dtbl.title
        ws_content[f'B{idx}'].hyperlink = f"#'{ws_dtbl.title}'!{str_cell_content_addr}"
        ws_content[f'B{idx}'].font = Font(color='0000FF')


    # def format_contribution_sheet(self, ws, ws_content):
    #
    #     self.format_content_sheet(ws_content, ws, str_cell_content_addr='A1')
    #
    #     ws.freeze_panes = 'C3'
    #
    #     ws.column_dimensions['A'].width = 20
    #     ws.column_dimensions['B'].width = 50
    #     ws.column_dimensions['C'].width = 18
    #
    #     for icol in range(1, ws.max_column + 1):
    #
    #         ws.cell(2, icol).font = Font(bold=True, size=12, color='FFFFFF')
    #         ws.cell(2, icol).fill = PatternFill(patternType='solid', fgColor='538DD5')
    #
    #         for irow in range(3, ws.max_row + 1):
    #             cur_cell = ws.cell(irow, icol)
    #             cur_cell.border = self.thin_border
    #             cur_cell.alignment = Alignment(horizontal='center', vertical='center')
    #
    #             if icol in [1, 2]:
    #                 cur_cell.font = Font(bold=True, size=12)
    #                 cur_cell.fill = PatternFill(patternType='solid', fgColor='D9E1F2')
    #             else:
    #                 cur_cell.font = Font(bold=False, size=12)
    #                 cur_cell.number_format = '0.0'
    #
    #
    # def format_brandmaster_sheet(self, ws, ws_content):
    #
    #     self.format_content_sheet(ws_content, ws, str_cell_content_addr='A1')
    #
    #     ws.freeze_panes = 'D3'
    #
    #     ws.column_dimensions['A'].width = 20
    #     ws.column_dimensions['B'].width = 50
    #     ws.column_dimensions['C'].width = 30
    #     ws.column_dimensions['D'].width = 30
    #     ws.column_dimensions['E'].width = 30
    #     ws.column_dimensions['F'].width = 30
    #
    #     ws.row_dimensions[2].height = 40
    #
    #     for icol in range(1, ws.max_column + 1):
    #
    #         ws.cell(2, icol).font = Font(bold=True, size=12, color='FFFFFF')
    #         ws.cell(2, icol).fill = PatternFill(patternType='solid', fgColor='538DD5')
    #         ws.cell(2, icol).alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
    #
    #         start_row = 3
    #         for irow in range(3, ws.max_row + 1):
    #             cur_cell = ws.cell(irow, icol)
    #             cur_cell.border = self.thin_border
    #             cur_cell.alignment = Alignment(horizontal='center', vertical='center')
    #
    #             if icol in [1]:
    #                 cur_cell.font = Font(bold=True, size=12)
    #                 cur_cell.fill = PatternFill(patternType='solid', fgColor='95B3D7')
    #             elif icol in [2]:
    #                 cur_cell.font = Font(bold=True, size=12)
    #                 cur_cell.fill = PatternFill(patternType='solid', fgColor='B8CCE4')
    #             elif icol in [3]:
    #                 cur_cell.font = Font(bold=True, size=12)
    #                 cur_cell.fill = PatternFill(patternType='solid', fgColor='DCE6F1')
    #
    #             else:
    #                 cur_cell.font = Font(bold=False, size=12)
    #                 cur_cell.number_format = '0.0'
    #
    #                 if icol == 6 and irow <= ws.max_row:
    #                     if ws.cell(irow, 3).value != ws.cell(irow + 1, 3).value:
    #                         ws.merge_cells(start_row=start_row, start_column=3, end_row=irow, end_column=3)
    #                         ws.merge_cells(start_row=start_row, start_column=6, end_row=irow, end_column=6)
    #                         start_row = irow + 1


    def format_side_axis(self, ws, is_matrix_table, start_side_row, num_format, step, lst_sub_header_col, is_tbl_sig=False):

        lst_sub_side_col = list()

        start_row = start_side_row

        for irow in range(start_row, ws.max_row + 1):

            cur1_cell = ws.cell(irow, 1)

            cur1_cell.border = self.thin_border
            cur1_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
            cur1_cell.font = Font(bold=True, color='FFFFFF')
            cur1_cell.fill = PatternFill(patternType='solid', fgColor='538DD5')

            cur2_cell = ws.cell(irow, 2)
            cur2_cell.border = self.thin_border
            cur2_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
            cur2_cell.font = Font(bold=True)
            cur2_cell.fill = PatternFill(patternType='solid', fgColor='D9E1F2')

            if cur2_cell.value != ws.cell(irow + 1, 2).value or cur1_cell.value != ws.cell(irow + 1, 1).value:

                if '_Group' in str(ws.cell(irow + 1, 2).value) or '_Mean' in str(ws.cell(irow + 1, 2).value):
                    continue

                end_row = irow
                ws.merge_cells(start_row=start_row, start_column=1, end_row=end_row, end_column=1)
                ws.merge_cells(start_row=start_row, start_column=2, end_row=end_row, end_column=2)
                start_row = irow + 1
                lst_sub_side_col.append(irow)

        if is_matrix_table:
            print(f"Formatting {ws.title} - Brand label")

            for icol in range(5, ws.max_column):
                for row in range(5, 7):
                    cur_cell = ws.cell(row, icol)
                    cur_cell.border = self.thin_border
                    cur_cell.font = Font(bold=True)

                    if row == 5:
                        cur_cell.alignment = Alignment(horizontal='center', vertical='center')
                        cur_cell.fill = PatternFill(patternType='solid', fgColor='D8E4BC')
                    elif row == 6:
                        cur_cell.fill = PatternFill(patternType='solid', fgColor='DAEEF3')

        start_row = start_side_row + 1 if is_matrix_table else start_side_row

        for irow in range(start_row, ws.max_row):

            if irow % 500 == 0:
                print(f"Formatting {ws.title} - row {irow}")

            if 'FT' in ws.cell(irow, 3).value:
                ws.row_dimensions[irow].hidden = True
            else:
                for icol in range(5, ws.max_column):
                    cur_cell = ws.cell(irow, icol)

                    cur_cell.border = self.dot_border

                    if icol in lst_sub_header_col:
                        cur_cell.border = self.dot_thin_right_border

                    if irow in lst_sub_side_col:
                        cur_cell.border = self.dot_thin_bot_border

                    if icol in lst_sub_header_col and irow in lst_sub_side_col:
                        cur_cell.border = self.dot_right_thin_right_border


                    if is_tbl_sig and ws.cell(start_side_row - 1, icol).value == 'A':
                        cur_cell.border = self.medium_left_1_border

                        if irow in lst_sub_side_col:
                            cur_cell.border = self.medium_left_2_border

                        if icol in lst_sub_header_col and irow in lst_sub_side_col:
                            cur_cell.border = self.medium_left_3_border

                        for irow_hd in range(3, start_side_row):
                            ws.cell(irow_hd, icol).border = self.medium_left_4_border


                    if ws.cell(irow, 4).value == 'base':
                        cur_cell.font = Font(bold=True)
                        cur_cell.fill = PatternFill(patternType='solid', fgColor='DAEEF3')

                    elif ws.cell(irow, 4).value == 'bes':
                        cur_cell.font = Font(bold=True)
                        cur_cell.number_format = '0.0'
                        cur_cell.fill = PatternFill(patternType='solid', fgColor='FDE9D9')

                        if icol == 6 or (icol > 6 and (icol - 6) % step == 0):
                            cur_cell.alignment = Alignment(horizontal='center', vertical='center')
                            ws.merge_cells(start_row=irow, start_column=icol, end_row=irow, end_column=icol + step - 1)

                    else:
                        if ws.cell(irow, 4).value in ['mean', 'std']:
                            ws.cell(irow, 5).font = Font(bold=True, color='0070C0')
                            cur_cell.font = Font(bold=True)
                            cur_cell.number_format = '0.00'

                        elif ws.cell(irow, 3).value == 'GROUP':
                            cur_cell.font = Font(bold=True)

                            if num_format == 'pct':
                                cur_cell.number_format = '0'
                            elif num_format == 'pct_sign':
                                cur_cell.number_format = '0%'

                        elif int(ws.cell(irow, 4).value) > 90000:

                            cur_cell.font = Font(bold=True)

                            if num_format == 'pct':
                                cur_cell.number_format = '0'
                            elif num_format == 'pct_sign':
                                cur_cell.number_format = '0%'

                            if '(NET)' in str(ws.cell(irow, 5).value).upper():
                                ws.cell(irow, 5).font = Font(bold=True, underline='double')
                            else:
                                ws.cell(irow, 5).font = Font(bold=True, underline='single')

                        else:
                            if icol > 5:
                                if num_format == 'pct':
                                    cur_cell.number_format = '0'
                                elif num_format == 'pct_sign':
                                    cur_cell.number_format = '0%'

                        if is_tbl_sig and icol % 2 != 0 and icol > 5:
                            cur_cell.font = Font(bold=True, color='FF0000')



    def format_sig_table(self):

        wb = openpyxl.load_workbook(self.xlsx_name, data_only=True)

        ws_content = wb['Content']
        ws_content.column_dimensions['B'].width = 60

        for ws in wb.worksheets:

            if ws.title in ['Content']:
                continue

            print(f"Formatting {ws.title}")

            self.format_content_sheet(ws_content, ws, str_cell_content_addr='A3')

            lst_sub_header_col = list()

            last_header_row = 2

            while ws.cell(last_header_row, 5).value != 'Base':
                last_header_row += 1

                if last_header_row >= 30:
                    last_header_row = -999
                    break

            last_header_row -= 1

            is_sig_tbl = True if ws.cell(4, 2).value else False

            if not is_sig_tbl:
                ws.delete_rows(last_header_row - 1, 2)
                last_header_row = last_header_row - 2
            else:
                ws.delete_rows(last_header_row - 1, 1)
                last_header_row = last_header_row - 1

            ws.freeze_panes = f'F{last_header_row + 1}'

            if 'count' in ws.cell(3, 2).value:
                num_format = 'count'
            else:
                if '%' in ws.cell(3, 2).value:
                    num_format = 'pct_sign'
                else:
                    num_format = 'pct'

            ws.column_dimensions['A'].width = 20
            ws.column_dimensions['B'].width = 40
            ws.column_dimensions['E'].width = 40

            if not is_sig_tbl:

                for icol in range(6, ws.max_column + 1).__reversed__():
                    if icol % 2 != 0:
                        ws.delete_cols(icol, 1)

            else:
                for icol in range(6, ws.max_column + 1):
                    ws.column_dimensions[get_column_letter(icol)].width = 7


            ws.column_dimensions['C'].hidden = True
            ws.column_dimensions['D'].hidden = True

            ws.row_dimensions[1].hidden = True
            ws.row_dimensions[2].hidden = True

            if is_sig_tbl and last_header_row > 1:
                ws.row_dimensions[last_header_row - 1].height = 30
            else:
                ws.row_dimensions[last_header_row].height = 30

            print(f"Formatting {ws.title} - Header")

            ws_max_col = ws.max_column

            dict_header_color = {
                0: {'fontColor': 'FFFFFF', 'fgColor': '203764'},
                1: {'fontColor': 'FFFFFF', 'fgColor': '305496'},
                2: {'fontColor': '000000', 'fgColor': '8EA9DB'},
                3: {'fontColor': '000000', 'fgColor': 'B4C6E7'},
                4: {'fontColor': '000000', 'fgColor': 'D9E1F2'},

                5: {'fontColor': '000000', 'fgColor': 'B5F1CD'},
                6: {'fontColor': '000000', 'fgColor': '57DF8E'},
                7: {'fontColor': '000000', 'fgColor': '27CF6B'},
                8: {'fontColor': 'FFFFFF', 'fgColor': '1E9E52'},
                9: {'fontColor': 'FFFFFF', 'fgColor': '156e3a'},

                # 5: {'fontColor': '000000', 'fgColor': 'A9D08E'},
                # 6: {'fontColor': '000000', 'fgColor': 'C6E0B4'},
                # 7: {'fontColor': '000000', 'fgColor': 'E2EFDA'},
                # 8: {'fontColor': '000000', 'fgColor': 'F4B084'},
                # 9: {'fontColor': '000000', 'fgColor': 'F8CBAD'},

                10: {'fontColor': '000000', 'fgColor': 'FCE4D6'},
                11: {'fontColor': '000000', 'fgColor': 'FCD5B4'},
                12: {'fontColor': '000000', 'fgColor': 'FDE9D9'},
            }

            for irow in range(3, last_header_row + 1):
                start_column = 6

                for icol in range(6, ws_max_col + 1):

                    cur_cell = ws.cell(irow, icol)

                    cur_cell.border = self.thin_border
                    cur_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
                    cur_cell.font = Font(bold=True, size=12, color=dict_header_color[irow - 3]['fontColor'])
                    cur_cell.fill = PatternFill(patternType='solid', fgColor=dict_header_color[irow - 3]['fgColor'])

                    if irow <= 4:

                        if cur_cell.value != ws.cell(irow, icol + 1).value:
                            end_column = icol

                            if start_column != end_column and cur_cell.value is not None:

                                ws.merge_cells(start_row=irow, start_column=start_column, end_row=irow, end_column=end_column)
                                lst_sub_header_col.append(icol)

                            start_column = icol + 1

                    else:

                        if cur_cell.value != ws.cell(irow, icol + 1).value or ws.cell(irow - 1, icol + 1).value is not None:
                            end_column = icol

                            if start_column != end_column and cur_cell.value is not None:

                                ws.merge_cells(start_row=irow, start_column=start_column, end_row=irow, end_column=end_column)
                                lst_sub_header_col.append(icol)

                            start_column = icol + 1


            print(f"Formatting {ws.title} - Side axis")



            start_side_row = last_header_row + 1
            self.format_side_axis(ws=ws, is_matrix_table=False, start_side_row=start_side_row, num_format=num_format,
                                  step=1, lst_sub_header_col=lst_sub_header_col, is_tbl_sig=is_sig_tbl)


        # output_name = self.xlsx_name.replace('.xlsx', '_output.xlsx')
        output_name = self.xlsx_name

        print(f"Save wb as {output_name}")
        wb.save(output_name)
        wb.close()


import os
import pandas as pd
import numpy as np
import time
import json
from scipy import stats



class DataTableGenerator:

    def __init__(self, df_data: pd.DataFrame, df_info: pd.DataFrame, xlsx_name: str, is_md: bool = False):

        self.df_data = df_data.copy()
        self.df_info = df_info.copy()

        self.is_md = is_md
        if is_md:
            self.convert_md_to_mc(df_data.copy(), df_info.copy())

        self.file_name = xlsx_name

        # lst_qre_group: list = [], lst_qre_mean: list = []
        # self.lst_qre_group = lst_qre_group
        # self.lst_qre_mean = lst_qre_mean

        self.dict_unnetted_qres = dict()
        for idx in self.df_info.index:
            if 'net_code' in self.df_info.at[idx, 'val_lbl'].keys():
                self.dict_unnetted_qres.update({self.df_info.at[idx, 'var_name']: self.unnetted_qre_val(self.df_info.at[idx, 'val_lbl'])})


    def convert_md_to_mc(self, df_data: pd.DataFrame, df_info: pd.DataFrame):

        df_info_qre_ma_all_1st_col = df_info.query("var_type.isin(['MA', 'MA_mtr']) & var_name.str.contains('[A-Za-z]+_1$')").copy()

        for idx in df_info_qre_ma_all_1st_col.index:
            qre = df_info_qre_ma_all_1st_col.at[idx, 'var_name'].rsplit('_', 1)[0]

            df_info_ma = df_info.query(f"var_name.str.contains('{qre}_[0-9]+')").copy()

            dict_val_lbl = dict()
            for idx_ma in df_info_ma.index:
                str_qre, str_val = str(df_info_ma.at[idx_ma, 'var_name']).rsplit('_', 1)
                dict_val_lbl[str_val] = df_info_ma.at[idx_ma, 'val_lbl']['1']

                df_data[df_info_ma.at[idx_ma, 'var_name']].replace({1: int(str_val)}, inplace=True)

            df_info.loc[df_info_ma.index,  ['val_lbl']] = [dict_val_lbl]

            lst_ma_col_name = df_info_ma['var_name'].values.tolist()

            df_data[lst_ma_col_name[0]] = df_data[lst_ma_col_name].values.tolist()

            for idx_dt in df_data.index:
                arr = np.sort(df_data.at[idx_dt, lst_ma_col_name[0]], axis=None)
                df_data.at[idx_dt, lst_ma_col_name[0]] = arr

            df_data[lst_ma_col_name] = pd.DataFrame(df_data[lst_ma_col_name[0]].tolist(), index=df_data.index)

        self.df_data = df_data
        self.df_info = df_info


    # def add_group(self):
    #     print('add_group')
    #     df_info = self.df_info.copy()
    #
    #     for qre_mean in self.lst_qre_group:
    #         idx = df_info.loc[df_info['var_name'] == qre_mean[0], :].index[0]
    #
    #         df_info = self.insert_row(idx + 1, df_info, [f'{qre_mean[0]}_Group', f"{df_info.at[idx, 'var_lbl']}_Group", 'GROUP', qre_mean[1]])
    #
    #     self.df_info = df_info
    #
    #
    # def add_mean(self):
    #     print('add_mean')
    #     df_info = self.df_info.copy()
    #
    #     for qre_mean in self.lst_qre_mean:
    #         idx = df_info.loc[df_info['var_name'] == qre_mean[0], :].index[0]
    #         df_info = self.insert_row(idx + 1, df_info, [f'{qre_mean[0]}_Mean',	f"{df_info.at[idx, 'var_lbl']}_Mean", 'MEAN', qre_mean[1]])
    #
    #     self.df_info = df_info


    def run_tables_by_js_files(self, lst_func_to_run: list, is_append: bool = False):

        # TODO: Update new way to run group and mean so these functions need to remove
        # self.add_group()
        # self.add_mean()

        if not is_append:
            if os.path.exists(self.file_name):
                os.remove(self.file_name)

            df_content_null = pd.DataFrame(columns=['#', 'Content'], data=[])

            with pd.ExcelWriter(self.file_name) as writer:
                df_content_null.to_excel(writer, sheet_name='Content', index=False)

        for item in lst_func_to_run:
            self.run_tables_by_item(item)


    def run_tables_by_item(self, item: dict):

        if 'json_file' in item.keys():

            with open(item['json_file'], encoding="UTF-8") as json_file:
                dict_tables = json.load(json_file)

        else:

            dict_tables = item['tables_format']

        if item['tables_to_run']:

            dict_tables_selected = dict()

            for tbl in item['tables_to_run']:
                dict_tables_selected[tbl] = dict_tables[tbl]

            dict_tables = dict_tables_selected

        for tbl_key, tbl_val in dict_tables.items():
            start_time = time.time()

            print(f"run table: {tbl_val['tbl_name']}")

            df_tbl = getattr(self, item['func_name'])(tbl_val)

            print(f"create sheet: {tbl_val['tbl_name']}")

            with pd.ExcelWriter(self.file_name, engine='openpyxl', mode='a', if_sheet_exists='overlay') as writer:
                df_tbl.to_excel(writer, sheet_name=tbl_val['tbl_name'], index=False)  # encoding='utf-8-sig'

            print(f"create sheet: {tbl_val['tbl_name']} in {time.time() - start_time} seconds")


    def group_sig_table_header(self, lst_header_qres: list) -> list:

        lst_group_header = list()

        while len(lst_header_qres) != 5:
            lst_header_qres.append([])

        lvl1, lvl2, lvl3, lvl4, lvl5 = lst_header_qres

        for a in lvl1:

            if not lvl2:
                dict_grp_hd = dict()
                dict_idx = {list(a['cats'].keys())[i]: i for i in range(len(a['cats'].keys()))}

                for a_k, a_v in a['cats'].items():
                    # qa = f"{a['qre_name']} > 0" if str(a_k).upper() == 'TOTAL' else (a_k if '@' in a['qre_name'] else f"{a['qre_name']} == {a_k}")
                    qa = self.group_sig_table_header_query(a, a_k)

                    dict_grp_hd.update({
                        dict_idx[a_k]: {
                            "lbl": f"{a['qre_lbl']}@{a_v}",
                            "query": f"{qa}"
                        }
                    })

                lst_group_header.append(dict_grp_hd)
                continue

            for a_k, a_v in a['cats'].items():
                # qa = f"{a['qre_name']} > 0" if str(a_k).upper() == 'TOTAL' else (a_k if '@' in a['qre_name'] else f"{a['qre_name']} == {a_k}")
                qa = self.group_sig_table_header_query(a, a_k)

                for b in lvl2:

                    if not lvl3:
                        dict_grp_hd = dict()
                        dict_idx = {list(b['cats'].keys())[i]: i for i in range(len(b['cats'].keys()))}

                        for b_k, b_v in b['cats'].items():
                            # qb = f"{b['qre_name']} > 0" if str(b_k).upper() == 'TOTAL' else f"{b['qre_name']} == {b_k}"
                            qb = self.group_sig_table_header_query(b, b_k)

                            dict_grp_hd.update({
                                dict_idx[b_k]: {
                                    "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}",
                                    "query": f"{qa} & {qb}"
                                }
                            })

                        lst_group_header.append(dict_grp_hd)
                        continue

                    for b_k, b_v in b['cats'].items():
                        # qb = f"{b['qre_name']} > 0" if str(b_k).upper() == 'TOTAL' else f"{b['qre_name']} == {b_k}"
                        qb = self.group_sig_table_header_query(b, b_k)

                        for c in lvl3:

                            if not lvl4:
                                dict_grp_hd = dict()
                                dict_idx = {list(c['cats'].keys())[i]: i for i in range(len(c['cats'].keys()))}

                                for c_k, c_v in c['cats'].items():
                                    # qc = f"{c['qre_name']} > 0" if str(c_k).upper() == 'TOTAL' else f"{c['qre_name']} == {c_k}"
                                    qc = self.group_sig_table_header_query(c, c_k)

                                    dict_grp_hd.update({
                                        dict_idx[c_k]: {
                                            "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}@{c['qre_lbl']}@{c_v}",
                                            "query": f"{qa} & {qb} & {qc}"
                                        }
                                    })

                                lst_group_header.append(dict_grp_hd)
                                continue

                            for c_k, c_v in c['cats'].items():
                                # qc = f"{c['qre_name']} > 0" if str(c_k).upper() == 'TOTAL' else f"{c['qre_name']} == {c_k}"
                                qc = self.group_sig_table_header_query(c, c_k)

                                for d in lvl4:

                                    if not lvl5:
                                        dict_grp_hd = dict()
                                        dict_idx = {list(d['cats'].keys())[i]: i for i in range(len(d['cats'].keys()))}

                                        for d_k, d_v in d['cats'].items():
                                            # qd = f"{d['qre_name']} > 0" if str(d_k).upper() == 'TOTAL' else f"{d['qre_name']} == {d_k}"
                                            qd = self.group_sig_table_header_query(d, d_k)

                                            dict_grp_hd.update({
                                                dict_idx[d_k]: {
                                                    "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}@{c['qre_lbl']}@{c_v}@{d['qre_lbl']}@{d_v}",
                                                    "query": f"{qa} & {qb} & {qc} & {qd}"
                                                }
                                            })

                                        lst_group_header.append(dict_grp_hd)
                                        continue

                                    for d_k, d_v in d['cats'].items():
                                        # qd = f"{d['qre_name']} > 0" if str(d_k).upper() == 'TOTAL' else f"{d['qre_name']} == {d_k}"
                                        qd = self.group_sig_table_header_query(d, d_k)

                                        for e in lvl5:
                                            dict_grp_hd = dict()
                                            dict_idx = {list(e['cats'].keys())[i]: i for i in range(len(e['cats'].keys()))}

                                            for e_k, e_v in e['cats'].items():
                                                # qe = f"{e['qre_name']} > 0" if str(e_k).upper() == 'TOTAL' else f"{e['qre_name']} == {e_k}"
                                                qe = self.group_sig_table_header_query(e, e_k)

                                                dict_grp_hd.update({
                                                    dict_idx[e_k]: {
                                                        "lbl": f"{a['qre_lbl']}@{a_v}@{b['qre_lbl']}@{b_v}@{c['qre_lbl']}@{c_v}@{d['qre_lbl']}@{d_v}@{e['qre_lbl']}@{e_v}",
                                                        "query": f"{qa} & {qb} & {qc} & {qd} & {qe}"
                                                    }
                                                })

                                            lst_group_header.append(dict_grp_hd)

        return lst_group_header


    def group_sig_table_header_query(self, dict_qre: dict, str_cat_or_query) -> str:

        if '$' in dict_qre['qre_name']:

            if 'RANK' in str(dict_qre['qre_name']).upper():
                lst_qre_ma_name = self.df_info.query(f"var_name.str.contains('{dict_qre['qre_name'][1:]}[0-9]+')")['var_name'].values.tolist()
            else:
                lst_qre_ma_name = self.df_info.query(f"var_name.str.contains('{dict_qre['qre_name'][1:]}_[0-9]+')")['var_name'].values.tolist()

            if str(str_cat_or_query).upper() == 'TOTAL':
                return f"({' | '.join([f'{i} > 0' for i in lst_qre_ma_name])})"

            return f"({' | '.join([f'{i} == {str_cat_or_query}' for i in lst_qre_ma_name])})"




        if str(str_cat_or_query).upper() == 'TOTAL':
            return f"{dict_qre['qre_name']} > 0"

        if '@' in dict_qre['qre_name']:
            return str_cat_or_query

        return f"{dict_qre['qre_name']} == {str_cat_or_query}"


    @staticmethod
    def unnetted_qre_val(dict_netted) -> dict:
        dict_unnetted = dict()

        if 'net_code' not in dict_netted.keys():
            return dict_netted

        for key, val in dict_netted.items():

            if 'net_code' in key:
                val_lbl_lv1 = dict_netted['net_code']

                for net_key, net_val in val_lbl_lv1.items():

                    if isinstance(net_val, str):
                        dict_unnetted.update({str(net_key): net_val})
                    else:
                        print(f"Unnetted {net_key}")
                        dict_unnetted.update(net_val)

            else:
                dict_unnetted.update({str(key): val})

        return dict_unnetted


    def run_standard_table_sig(self, tbl: dict) -> pd.DataFrame:

        df_tbl = pd.DataFrame()

        # create df_data with tbl_filter in json file
        df_data = self.df_data.query(tbl['tbl_filter']).copy()

        # create df_info with lst_side_qres in json file
        df_info = pd.DataFrame(columns=['var_name', 'var_lbl', 'var_type', 'val_lbl', 'qre_fil'], data=[])
        for qre in tbl['lst_side_qres']:

            if '$' in qre['qre_name']:

                if '_RANK' in str(qre['qre_name']).upper():
                    lst_qre_col = self.df_info.loc[self.df_info['var_name'].str.contains(f"{qre['qre_name'][1:]}[0-9]+"), 'var_name'].values.tolist()
                else:
                    lst_qre_col = self.df_info.loc[self.df_info['var_name'].str.contains(f"{qre['qre_name'][1:]}_[0-9]+"), 'var_name'].values.tolist()

                var_name = qre['qre_name'].replace('$', '')

            elif '#combine' in qre['qre_name']:
                var_name, str_comb = qre['qre_name'].split('#combine')
                lst_qre_col = str_comb.replace('(', '').replace(')', '').split(',')
            else:
                lst_qre_col = [qre['qre_name']]
                var_name = qre['qre_name']

            # NEW-------------------------------------------------------------------------------------------------------
            df_qre_info = self.df_info.query(f"var_name.isin({lst_qre_col})").copy()
            df_qre_info.reset_index(drop=True, inplace=True)

            dict_row = {
                'var_name': var_name,
                'var_lbl': qre['qre_lbl'].replace('{lbl}', df_qre_info.at[0, 'var_lbl']) if 'qre_lbl' in qre.keys() else df_qre_info.at[0, 'var_lbl'],
                'var_type': 'MA_comb' if '#combine' in qre['qre_name'] else ('MA_Rank' if '$' in qre['qre_name'] and '_RANK' in str(qre['qre_name']).upper() else df_qre_info.at[0, 'var_type']),
                'val_lbl': qre['cats'] if 'cats' in qre.keys() else df_qre_info.at[0, 'val_lbl'],
                'qre_fil': qre['qre_filter'] if 'qre_filter' in qre.keys() else "",
                'lst_qre_col': lst_qre_col,
                'mean': qre['mean'] if 'mean' in qre.keys() else {},
                'sort': qre['sort'] if 'sort' in qre.keys() else ''
            }

            df_info = pd.concat([df_info, pd.DataFrame(columns=list(dict_row.keys()), data=[list(dict_row.values())])], axis=0, ignore_index=True)
            # ----------------------------------------------------------------------------------------------------------


        # Maximum 5 levels of header
        lst_group_header = self.group_sig_table_header(tbl['lst_header_qres'])

        for grp_hd in lst_group_header:

            print(f"run table: {tbl['tbl_name']} -> group header:")

            for i in grp_hd.values():
                print(f"\t{i['lbl']}")

            tbl_info_sig = {
                'tbl_name': tbl['tbl_name'],
                'is_count': tbl['is_count'],
                'is_pct_sign': tbl['is_pct_sign'],
                'sig_test_info': tbl['sig_test_info'],
                'dict_grp_header': grp_hd
            }

            df_temp = self.run_standard_header_sig(df_data, df_info, tbl_info_sig=tbl_info_sig)

            if df_tbl.empty:
                df_tbl = df_temp
            else:
                lst_col_temp_to_add = list(df_temp.columns)[5:]
                df_tbl = pd.concat([df_tbl, df_temp[lst_col_temp_to_add]], axis=1)


        # drop row which have all value is nan
        df_tbl.dropna(how='all', inplace=True)

        # Drop rows in qre oe that have all columns are 0
        if tbl['is_hide_oe_zero_cats']:

            df_sum_oe_val = df_tbl.query("qre_name.str.contains('_OE') & qre_type == 'MA'").copy()

            if not df_sum_oe_val.empty:
                fil_col = list(df_sum_oe_val.columns)
                df_sum_oe_val = df_sum_oe_val.loc[:, fil_col[5:]]
                df_sum_oe_val.replace({'': np.nan}, inplace=True)
                df_sum_oe_val = df_sum_oe_val.astype(float)
                df_sum_oe_val['sum_val'] = df_sum_oe_val.sum(axis=1, skipna=True, numeric_only=True)
                df_sum_oe_val = df_sum_oe_val.query('sum_val == 0')

                df_tbl.drop(df_sum_oe_val.index, inplace=True)


        # Drop columns which all value equal 0
        if tbl['is_hide_zero_cols']:

            start_idx = df_tbl.query(f"cat_val == 'base'").index.tolist()[0]
            lst_val_col = [v for i, v in enumerate(df_tbl.columns.tolist()[5:]) if i % 2 == 0]

            df_fil = df_tbl.query("index >= @start_idx")[lst_val_col].copy()
            df_fil.replace({0: np.nan}, inplace=True)
            df_fil.dropna(axis='columns', how='all', inplace=True)

            lst_keep_col = list()
            for i in df_fil.columns.tolist():
                lst_keep_col.extend([i, i.replace('@val@', '@sig@')])

            df_tbl = df_tbl[df_tbl.columns.tolist()[:5] + lst_keep_col]

            # df_tbl.to_excel('df_tbl_review.xlsx')
            a = 1



        # Reset df table index
        df_tbl.reset_index(drop=True, inplace=True)

        # df_tbl.to_excel('zzz_df_tbl.xlsx', encoding='utf-8-sig')
        
        return df_tbl


    @staticmethod
    def add_base_to_tbl_sig(df_data: pd.DataFrame, df_tbl: pd.DataFrame, df_qre: pd.DataFrame, qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list) -> pd.DataFrame:

        lst_tbl_row_data = list()
        lst_ignore_col = list()

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            for idx_item, item in enumerate(sig_pair):

                if item in lst_ignore_col:
                    continue

                str_hd_query = df_tbl.at[0, dict_header_col_name[item]['val_col']]

                if len(qre_info['lst_qre_col']) > 1:
                    lst_qre_col = qre_info['lst_qre_col']
                    df_filter = df_data.query(str_hd_query).copy()
                    df_fil_base = df_filter[lst_qre_col].dropna(how='all')
                    df_filter = df_filter.loc[df_fil_base.index, :]

                else:
                    str_query = f"{str_hd_query} & {qre_info['qre_name'].replace('_Mean', '').replace('_Group', '')} > 0"
                    df_filter = df_data.query(str_query).copy()

                if df_data.empty:
                    num_base = 0
                else:
                    num_base = df_filter.shape[0]

                if len(lst_tbl_row_data) == 0:
                    # str_qre_name = qre_info['qre_name'].rsplit('_', 1)[0] if qre_info['qre_type'] in ['MA', 'MA_mtr', 'MA_comb'] else qre_info['qre_name']
                    str_qre_name = qre_info['qre_name']
                    lst_tbl_row_data = [str_qre_name, qre_info['qre_lbl'], qre_info['qre_type'], 'base', 'Base', num_base, np.nan]
                else:
                    lst_tbl_row_data.extend([num_base, np.nan])

                lst_ignore_col.append(item)

        df_qre.loc[len(df_qre)] = lst_tbl_row_data
        return df_qre


    def add_sa_qre_val_to_tbl_sig(self, df_data: pd.DataFrame, df_tbl: pd.DataFrame, df_qre: pd.DataFrame,
                                  qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list,
                                  sig_type: str, lst_sig_lvl: list, cat: str, lbl: str, lst_sub_cat: list = None) -> pd.DataFrame:

        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val']
        is_count = qre_info['is_count']
        val_pct = qre_info['val_pct']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_name': qre_name,
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cat,
            'cat_lbl': lbl,
        })

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):
                str_query = f"{df_tbl.at[0, dict_header_col_name[item]['val_col']]}"

                df_filter = df_data.query(str_query).loc[:, [qre_name]].copy()

                if df_filter.empty:
                    continue

                if lst_sub_cat:
                    dict_re_qre_val = {int(k): 1 if k in lst_sub_cat else 0 for k, v in qre_val.items()}
                else:
                    dict_re_qre_val = {int(k): 1 if k == cat else 0 for k, v in qre_val.items()}

                df_filter.replace(dict_re_qre_val, inplace=True)

                dict_pair_to_sig.update({item: df_filter})

                num_val = (df_filter[qre_name] == 1).sum() if is_count else df_filter[qre_name].mean() * val_pct

                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:
                    num_val_old = df_qre.loc[df_qre['cat_val'] == cat, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]


            if sig_type and lst_sig_lvl:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre


    def add_sa_qre_mean_to_tbl_sig(self, df_data: pd.DataFrame, df_tbl: pd.DataFrame, df_qre: pd.DataFrame,
                                   qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list,
                                   sig_type: str, lst_sig_lvl: list, mean_factor: dict, is_mean: bool = True) -> pd.DataFrame:

        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val']

        if is_mean:
            dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
            dict_new_row.update({
                'qre_name': qre_name,
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'cat_val': 'mean',
                'cat_lbl': 'Mean',
            })
        else:
            dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
            dict_new_row.update({
                'qre_name': qre_name.replace('_Mean', '_Std'),
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'cat_val': 'std',
                'cat_lbl': 'Std',
            })

        org_qre_name = qre_name.replace('_Mean', '')

        if mean_factor:
            df_data.replace(mean_factor, inplace=True)

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):
                str_query = f"{df_tbl.at[0, dict_header_col_name[item]['val_col']]}"

                df_filter = df_data.query(str_query).loc[:, [org_qre_name]].copy()

                if df_filter.empty:
                    continue

                if -999 not in qre_val.keys():
                    df_filter.replace(qre_val, inplace=True)

                dict_pair_to_sig.update({item: df_filter})

                if is_mean:

                    num_val = df_filter[org_qre_name].mean()
                    val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                    if sig_type and lst_sig_lvl:

                        num_val_old = df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name]].values[0, 0]

                        if pd.isnull(num_val_old):
                            df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name, sig_col_name]] = [num_val, np.nan]

                    else:
                        df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    num_val_std = df_filter[org_qre_name].std()
                    val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                    df_qre.loc[df_qre['cat_val'] == 'std', [val_col_name, sig_col_name]] = [num_val_std, np.nan]


            if sig_type and lst_sig_lvl and is_mean:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre


    def add_sa_qre_group_to_tbl_sig(self, df_data: pd.DataFrame, df_tbl: pd.DataFrame, df_qre: pd.DataFrame,
                                    qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list,
                                    sig_type: str, lst_sig_lvl: list, cat: str, lbl: str) -> pd.DataFrame:  # lst_sub_cat: list = None

        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val']
        is_count = qre_info['is_count']
        val_pct = qre_info['val_pct']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_name': qre_name,
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cat,
            'cat_lbl': lbl,
        })

        org_qre_name = qre_name.replace('_Group', '')

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):
                str_query = f"{df_tbl.at[0, dict_header_col_name[item]['val_col']]}"

                df_filter = df_data.query(str_query).loc[:, [org_qre_name]].copy()

                if df_filter.empty:
                    continue

                df_filter.replace(qre_val['recode'], inplace=True)

                dict_re_qre_val = {int(k): 1 if int(k) == int(cat) else 0 for k, v in qre_val['cats'].items()}
                df_filter.replace(dict_re_qre_val, inplace=True)

                dict_pair_to_sig.update({item: df_filter})

                num_val = (df_filter[org_qre_name] == 1).sum() if is_count else df_filter[org_qre_name].mean() * val_pct
                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:

                    num_val_old = df_qre.loc[df_qre['cat_val'] == cat, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]


            if sig_type and lst_sig_lvl:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre


    def add_num_qre_to_tbl_sig(self, df_data: pd.DataFrame, df_tbl: pd.DataFrame, df_qre: pd.DataFrame,
                               qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list,
                               sig_type: str, lst_sig_lvl: list) -> pd.DataFrame:

        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_name': qre_name,
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': 'mean',
            'cat_lbl': 'Mean',
        })

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):
                str_query = f"{df_tbl.at[0, dict_header_col_name[item]['val_col']]}"
                df_filter = df_data.query(str_query).loc[:, [qre_name]].copy()

                if df_filter.empty:
                    continue

                dict_pair_to_sig.update({item: df_filter})

                num_val = df_filter[qre_name].mean()
                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:

                    num_val_old = df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:

                    df_qre.loc[df_qre['cat_val'] == 'mean', [val_col_name, sig_col_name]] = [num_val, np.nan]

            if sig_type and lst_sig_lvl:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre


    def add_ma_qre_val_to_tbl_sig(self, df_data: pd.DataFrame, df_tbl: pd.DataFrame, df_qre: pd.DataFrame,
                                  qre_info: dict, dict_header_col_name: dict, lst_sig_pair: list,
                                  sig_type: str, lst_sig_lvl: list, cat: str, lbl: str, lst_sub_cat: list = None) -> pd.DataFrame:

        if lst_sub_cat is None:
            lst_sub_cat = []

        qre_name = qre_info['qre_name']
        qre_lbl = qre_info['qre_lbl']
        qre_type = qre_info['qre_type']
        qre_val = qre_info['qre_val']
        is_count = qre_info['is_count']
        val_pct = qre_info['val_pct']
        # df_ma_info = qre_info['df_ma_info']
        lst_qre_col = qre_info['lst_qre_col']

        dict_new_row = {col: '' if '@sig@' in col else np.nan for col in df_qre.columns}
        dict_new_row.update({
            'qre_name': qre_name,  # qre_name.rsplit('_', 1)[0],
            'qre_lbl': qre_lbl,
            'qre_type': qre_type,
            'cat_val': cat,
            'cat_lbl': lbl,
        })

        df_qre = pd.concat([df_qre, pd.DataFrame(columns=list(dict_new_row.keys()), data=[list(dict_new_row.values())])], axis=0, ignore_index=True)

        for idx_pair, sig_pair in enumerate(lst_sig_pair):

            dict_pair_to_sig = dict()

            for idx_item, item in enumerate(sig_pair):
                str_query = f"{df_tbl.at[0, dict_header_col_name[item]['val_col']]}"
                # df_filter = df_data.query(str_query).loc[:, df_ma_info['var_name'].values.tolist()].copy()
                df_filter = df_data.query(str_query).loc[:, lst_qre_col].copy()

                if df_filter.empty:
                    continue

                if lst_sub_cat:
                    dict_re_qre_val = {int(k): 1 if k in lst_sub_cat else 0 for k, v in qre_val.items()}

                else:
                    dict_re_qre_val = {int(k): 1 if k == cat else 0 for k, v in qre_val.items()}

                df_filter.replace(dict_re_qre_val, inplace=True)

                # df_filter['ma_val_sum'] = df_filter.loc[df_filter[qre_name] >= 0, :].sum(axis='columns')

                df_fil_base = df_filter[lst_qre_col].dropna(how='all')
                df_filter.loc[df_fil_base.index, 'ma_val_sum'] = df_filter.loc[df_fil_base.index, lst_qre_col].sum(axis='columns')


                if lst_sub_cat or qre_type == 'MA_comb':
                    df_filter.loc[df_filter['ma_val_sum'] > 1, 'ma_val_sum'] = 1

                dict_pair_to_sig.update({item: df_filter['ma_val_sum']})

                num_val = (df_filter['ma_val_sum'] == 1).sum() if is_count else df_filter['ma_val_sum'].mean() * val_pct
                val_col_name, sig_col_name = dict_header_col_name[item]['val_col'], dict_header_col_name[item]['sig_col']

                if sig_type and lst_sig_lvl:
                    num_val_old = df_qre.loc[df_qre['cat_val'] == cat, [val_col_name]].values[0, 0]

                    if pd.isnull(num_val_old):
                        df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

                else:
                    df_qre.loc[df_qre['cat_val'] == cat, [val_col_name, sig_col_name]] = [num_val, np.nan]

            if sig_type and lst_sig_lvl:
                df_qre = self.mark_sig_to_df_qre(df_qre, dict_pair_to_sig, sig_pair, dict_header_col_name, sig_type, lst_sig_lvl)

        return df_qre


    @staticmethod
    def mark_sig_to_df_qre(df_qre: pd.DataFrame, dict_pair_to_sig: dict, sig_pair: list, dict_header_col_name: dict, sig_type: str, lst_sig_lvl: list) -> pd.DataFrame:

        if not lst_sig_lvl or not sig_type or not dict_pair_to_sig:
            return df_qre

        if sig_pair[0] not in dict_pair_to_sig.keys() or sig_pair[1] not in dict_pair_to_sig.keys():
            return df_qre

        df_left, df_right = dict_pair_to_sig[sig_pair[0]], dict_pair_to_sig[sig_pair[1]]

        if df_left.shape[0] < 30 or df_right.shape[0] < 30:
            return df_qre

        is_df_left_null = df_left.isnull().values.all()
        is_df_right_null = df_right.isnull().values.all()

        if is_df_left_null or is_df_right_null:
            return df_qre

        try:
            if df_left.mean().iloc[0] == 0 or df_right.mean().iloc[0] == 0:
                return df_qre

            if df_left.mean().iloc[0] == 1 and df_right.mean().iloc[0] == 1:
                return df_qre

        except Exception:
            if df_left.mean() == 0 or df_right.mean() == 0:
                return df_qre

            if df_left.mean() == 1 and df_right.mean() == 1:
                return df_qre

        if sig_type == 'rel':
            if df_left.shape[0] != df_right.shape[0]:
                return df_qre

            sigResult = stats.ttest_rel(df_left, df_right)
        else:
            sigResult = stats.ttest_ind_from_stats(
                mean1=df_left.mean(), std1=df_left.std(), nobs1=df_left.shape[0],
                mean2=df_right.mean(), std2=df_right.std(), nobs2=df_right.shape[0]
            )

        if sigResult.pvalue:
            if sigResult.statistic > 0:
                mark_sig_char = sig_pair[1]
                sig_col_name = dict_header_col_name[sig_pair[0]]['sig_col']
            else:
                mark_sig_char = sig_pair[0]
                sig_col_name = dict_header_col_name[sig_pair[1]]['sig_col']

            if sigResult.pvalue <= lst_sig_lvl[0]:
                df_qre.at[df_qre.index[-1], sig_col_name] = str(df_qre.at[df_qre.index[-1], sig_col_name]).replace('nan', '') + mark_sig_char.upper()
            elif len(lst_sig_lvl) >= 2:
                if sigResult.pvalue <= lst_sig_lvl[1]:
                    df_qre.loc[df_qre.index[-1], sig_col_name] = str(df_qre.at[df_qre.index[-1], sig_col_name]).replace('nan', '') + mark_sig_char.lower()

        return df_qre



    def run_standard_header_sig(self, df_data: pd.DataFrame, df_info: pd.DataFrame, tbl_info_sig: dict) -> pd.DataFrame:

        is_count = tbl_info_sig['is_count']
        val_pct = 1 if tbl_info_sig['is_pct_sign'] else 100

        sig_type = tbl_info_sig['sig_test_info']['sig_type']

        lst_sig_lvl_pct = tbl_info_sig['sig_test_info']['lst_sig_lvl']
        lst_sig_lvl = [(100 - a) / 100 for a in lst_sig_lvl_pct]
        lst_sig_lvl.reverse()

        dict_grp_header = tbl_info_sig['dict_grp_header']

        dict_char_sig = {
            0: 'A',
            1: 'B',
            2: 'C',
            3: 'D',
            4: 'E',
            5: 'F',
            6: 'G',
            7: 'H',
            8: 'I',
            9: 'J',
            10: 'K',
            11: 'L',
            12: 'M',
            13: 'N',
            14: 'O',
            15: 'P',
            16: 'Q',
            17: 'R',
            18: 'S',
            19: 'T',
            20: 'U',
            21: 'V',
            22: 'W',
            23: 'X',
            24: 'Y',

        }

        # lst_tbl_col = ['qre_name', 'qre_lbl', 'qre_type', 'cat_val', 'cat_lbl']
        dict_tbl_data = {
            'qre_name': list(),
            'qre_lbl': list(),
            'qre_type': list(),
            'cat_val': list(),
            'cat_lbl': list(),
        }

        dict_header_col_name = dict()

        for hd_k, hd_v in dict_grp_header.items():
            str_hd_val = f"{hd_v['query']}@{hd_v['lbl']}@val@{dict_char_sig[hd_k]}"
            str_hd_sig = f"{hd_v['query']}@{hd_v['lbl']}@sig@{dict_char_sig[hd_k]}"

            dict_tbl_data.update({
                str_hd_val: str_hd_val.split('@'),
                str_hd_sig: str_hd_sig.split('@'),
            })

            dict_header_col_name.update({
                dict_char_sig[hd_k]: {
                    'val_col': str_hd_val,
                    'sig_col': str_hd_sig,
                }
            })

            if len(dict_tbl_data['qre_name']) == 0:
                arr_nan = [np.nan] * len(str_hd_val.split('@'))
                dict_tbl_data.update({
                    'qre_name': arr_nan,
                    'qre_lbl': arr_nan,
                    'qre_type': arr_nan,
                    'cat_val': arr_nan,
                    'cat_lbl': arr_nan,
                })

        if sig_type in ["ind", "rel"]:

            if tbl_info_sig['sig_test_info']['sig_cols']:
                lst_sig_pair = tbl_info_sig['sig_test_info']['sig_cols']
            else:
                lst_sig_char = list(dict_header_col_name.keys())
                lst_sig_pair = list()
                for i in range(len(lst_sig_char)-1):
                    for j in range(i + 1, len(lst_sig_char)):
                        lst_sig_pair.append([lst_sig_char[i], lst_sig_char[j]])

        else:
            lst_sig_char = list(dict_header_col_name.keys())
            lst_sig_pair = list()
            # for j in range(1, len(lst_sig_char)):
            #     lst_sig_pair.append([lst_sig_char[0], lst_sig_char[j]])

            for i in lst_sig_char:
                lst_sig_pair.append([i])


        df_tbl = pd.DataFrame.from_dict(dict_tbl_data)

        df_tbl['qre_lbl'] = df_tbl['qre_lbl'].astype('object')

        df_tbl.loc[1:4, ['qre_lbl']] = [
            f"Cell content: {'count' if is_count else ('percentage(%)' if tbl_info_sig['is_pct_sign'] else 'percentage')}",
            '' if is_count or sig_type == '' else f"{'Dependent' if sig_type == 'rel' else 'Independent'} Pair T-test at level {' & '.join([f'{i}%' for i in lst_sig_lvl_pct])}",
            '' if is_count or sig_type == '' else f"Columns Tested: {', '.join(['/'.join(i) for i in lst_sig_pair])}",
            '' if is_count or sig_type == '' else f"Uppercase for {lst_sig_lvl_pct[-1]}%, lowercase for {lst_sig_lvl_pct[0]}%" if len(lst_sig_lvl_pct) > 1 else np.nan
        ]

        # lst_ignore_col_name = list()

        for idx in df_info.index:

            qre_name = df_info.at[idx, 'var_name']
            qre_lbl = df_info.at[idx, 'var_lbl']
            qre_type = df_info.at[idx, 'var_type']
            qre_val = eval(df_info.at[idx, 'val_lbl']) if isinstance(df_info.at[idx, 'val_lbl'], str) else df_info.at[idx, 'val_lbl']
            qre_fil = df_info.at[idx, 'qre_fil']
            lst_qre_col = df_info.at[idx, 'lst_qre_col']


            print(qre_name)

            # if qre_name == 'Q2_OL_Com':
            #     a = 1

            # filter data base on qre_fil
            df_data_qre_fil = df_data.query(qre_fil).copy() if qre_fil else df_data.copy()

            qre_info = {
                'qre_name': qre_name,
                'qre_lbl': qre_lbl,
                'qre_type': qre_type,
                'qre_val': qre_val,
                'is_count': is_count,
                'val_pct': val_pct,
                'lst_qre_col': lst_qre_col,
            }

            df_qre = pd.DataFrame(columns=df_tbl.columns, data=[])

            # qre_name_ma, qre_ma_col_order = 0, 0
            # if qre_type in ['MA', 'MA_mtr', 'MA_comb']:
            #     qre_name_ma, qre_ma_col_order = qre_name.rsplit('_', 1)
            #     qre_ma_col_order = int(qre_ma_col_order)

            # BASE------------------------------------------------------------------------------------------------------
            # if qre_type in ['MEAN', 'GROUP'] or (qre_type in ['MA', 'MA_mtr', 'MA_comb'] and qre_ma_col_order > 1):
            if qre_type in ['MEAN', 'GROUP']:
                pass
            else:
                df_qre = self.add_base_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair)

            # END BASE--------------------------------------------------------------------------------------------------

            if qre_type in ['FT', 'FT_mtr']:
                # Not run free text questions
                pass

            elif qre_type in ['SA', 'SA_mtr', 'RANKING']:

                if qre_name in self.dict_unnetted_qres.keys():
                    qre_val_unnetted = self.dict_unnetted_qres[qre_name]
                else:
                    if 'net_code' in qre_val.keys():
                        self.dict_unnetted_qres.update({qre_name: self.unnetted_qre_val(qre_val)})
                        qre_val_unnetted = self.dict_unnetted_qres[qre_name]
                    else:
                        qre_val_unnetted = qre_val

                qre_info['qre_val'] = qre_val_unnetted

                for cat, lbl in qre_val.items():

                    # ADD-IN Net Code
                    if 'net_code' in str(cat):

                        for net_cat, net_val in qre_val['net_code'].items():

                            if isinstance(net_val, dict):

                                lst_sub_cat = list(net_val.keys())

                                net_cat_val, net_cat_type, net_cat_lbl = net_cat.split('|')

                                df_qre = self.add_sa_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat_val, net_cat_lbl, lst_sub_cat)

                                if 'NET' in net_cat_type.upper():
                                    for cat2, lbl2 in net_val.items():
                                        df_qre = self.add_sa_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat2, f' - {lbl2}')
                            else:
                                df_qre = self.add_sa_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat, net_val)

                    else:
                        df_qre = self.add_sa_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat, lbl)

                mean_factor = df_info.at[idx, 'mean']
                if mean_factor.keys():
                    # Run Mean
                    df_qre = self.add_sa_qre_mean_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, mean_factor)

                    # Run Std
                    df_qre = self.add_sa_qre_mean_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, mean_factor, is_mean=False)

            # elif qre_type in ['MEAN']:
            #
            #     df_qre = self.add_sa_qre_mean_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl)
            #
            # elif qre_type in ['GROUP']:
            #
            #     for cat, lbl in qre_val['cats'].items():
            #         df_qre = self.add_sa_qre_group_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat, lbl)

            elif qre_type in ['NUM']:

                df_qre = self.add_num_qre_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl)

            elif qre_type in ['MA', 'MA_mtr', 'MA_comb', 'MA_Rank']:

                # if 'Q1a' in qre_name:
                #     a = 1

                if f'{qre_name}_1' in self.dict_unnetted_qres.keys() and f'{qre_name}_2' in self.dict_unnetted_qres.keys():
                    qre_val_unnetted = self.dict_unnetted_qres[f'{qre_name}_1']
                else:
                    if f'{qre_name}_1' in self.dict_unnetted_qres.keys():
                        qre_val_unnetted = self.dict_unnetted_qres[f'{qre_name}_1']
                    else:
                        if 'net_code' in qre_val.keys():
                            self.dict_unnetted_qres.update({f'{qre_name}_1': self.unnetted_qre_val(qre_val)})
                            qre_val_unnetted = self.dict_unnetted_qres[f'{qre_name}_1']
                        else:
                            qre_val_unnetted = qre_val

                qre_info['qre_val'] = qre_val_unnetted

                for cat, lbl in qre_val.items():

                    # ADD-IN Net Code
                    if 'net_code' in cat:

                        for net_cat, net_val in qre_val['net_code'].items():

                            if isinstance(net_val, dict):

                                lst_sub_cat = list(net_val.keys())

                                # "900001|combine|POSITIVE (NET)"
                                # list_net_cat = net_cat.split('|')
                                net_cat_val, net_cat_type, net_cat_lbl = net_cat.split('|')

                                df_qre = self.add_ma_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat_val, net_cat_lbl, lst_sub_cat)

                                if 'NET' in net_cat_type.upper():
                                    for cat2, lbl2 in net_val.items():
                                        df_qre = self.add_ma_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat2, f' - {lbl2}')

                            else:
                                df_qre = self.add_ma_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, net_cat, net_val)

                    else:
                        df_qre = self.add_ma_qre_val_to_tbl_sig(df_data_qre_fil, df_tbl, df_qre, qre_info, dict_header_col_name, lst_sig_pair, sig_type, lst_sig_lvl, cat, lbl)


            # SORTING---------------------------------------------------------------------------------------------------
            sort_opt = df_info.at[idx, 'sort']

            if sort_opt:
                is_asc = True if sort_opt == 'asc' else False
                base_val = -999_999_999 if sort_opt == 'asc' else 999_999_999

                df_qre['sort_col'] = df_qre[df_qre.columns.tolist()[5]]
                df_qre.loc[df_qre['cat_val'] == 'base', 'sort_col'] = base_val

                df_qre.sort_values(by=['sort_col'], ascending=is_asc, inplace=True, ignore_index=True)
                df_qre.drop(columns=['sort_col'], inplace=True)

            # END SORTING-----------------------------------------------------------------------------------------------

            df_tbl = pd.concat([df_tbl, df_qre], axis=0, ignore_index=True)

        return df_tbl